@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2015. All Rights Reserved.
+%% Copyright Ericsson AB 2015-2016. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -44,11 +44,11 @@
 %% This depends on the algorithm handler function
 -type alg_seed() :: exs64_state() | exsplus_state() | exs1024_state().
 %% This is the algorithm handler function within this module
--type alg_handler() :: #{type      => alg(),
-			 max       => integer(),
-			 next      => fun(),
-			 uniform   => fun(),
-			 uniform_n => fun()}.
+-type alg_handler() :: #{type      := alg(),
+			 max       := integer(),
+			 next      := fun(),
+			 uniform   := fun(),
+			 uniform_n := fun()}.
 
 %% Internal state
 -opaque state() :: {alg_handler(), alg_seed()}.
@@ -63,7 +63,7 @@
 %% Return algorithm and seed so that RNG state can be recreated with seed/1
 -spec export_seed() -> undefined | export_state().
 export_seed() ->
-    case seed_get() of
+    case get(?SEED_DICT) of
 	{#{type:=Alg}, Seed} -> {Alg, Seed};
 	_ -> undefined
     end.
@@ -256,6 +256,8 @@ exs64_uniform(Max, {Alg, R}) ->
 %% =====================================================================
 -type exsplus_state() :: nonempty_improper_list(uint58(), uint58()).
 
+-dialyzer({no_improper_lists, exsplus_seed/1}).
+
 exsplus_seed({A1, A2, A3}) ->
     {_, R1} = exsplus_next([(((A1 * 4294967197) + 1) band ?UINT58MASK)|
 			    (((A2 * 4294967231) + 1) band ?UINT58MASK)]),
@@ -263,6 +265,8 @@ exsplus_seed({A1, A2, A3}) ->
 			    tl(R1)]),
     R2.
 
+-dialyzer({no_improper_lists, exsplus_next/1}).
+
 %% Advance xorshift116+ state for one step and generate 58bit unsigned integer
 -spec exsplus_next(exsplus_state()) -> {uint58(), exsplus_state()}.
 exsplus_next([S1|S0]) ->
